"""
Helpers to support streaming generate output.
Borrowed from https://github.com/oobabooga/text-generation-webui/blob/ad37f396fc8bcbab90e11ecf17c56c97bfbd4a9c/modules/callbacks.py
"""

import gc
import traceback
from queue import Queue
from threading import Thread
import threading
import os
import requests

import torch
import transformers
from transformers import TrainerCallback

class Stream(transformers.StoppingCriteria):
    def __init__(self, callback_func=None):
        self.callback_func = callback_func

    def __call__(self, input_ids, scores) -> bool:
        if self.callback_func is not None:
            self.callback_func(input_ids[0])
        return False


class Iteratorize:

    """
    Transforms a function that takes a callback
    into a lazy iterator (generator).
    """

    def __init__(self, func, kwargs={}, callback=None):
        self.mfunc = func
        self.c_callback = callback
        self.q = Queue()
        self.sentinel = object()
        self.kwargs = kwargs
        self.stop_now = False

        def _callback(val):
            if self.stop_now:
                raise ValueError
            self.q.put(val)

        def gentask():
            try:
                ret = self.mfunc(callback=_callback, **self.kwargs)
            except ValueError:
                pass
            except:
                traceback.print_exc()
                pass

            self.q.put(self.sentinel)
            if self.c_callback:
                self.c_callback(ret)

        self.thread = Thread(target=gentask)
        self.thread.start()

    def __iter__(self):
        return self

    def __next__(self):
        obj = self.q.get(True, None)
        if obj is self.sentinel:
            raise StopIteration
        else:
            return obj

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_now = True

import time
class TrainProgress(TrainerCallback):
    def __init__(self, cb):
        self.start_time = time.time()
        self.cb = cb

    def on_train_begin(self, args, state, control, **kwargs):
        self.start_time = time.time()

    def on_step_end(self, args, state, control, **kwargs):
        cur_step = state.global_step
        max_step = state.max_steps
        process : float = cur_step * 100 // max_step

        if cur_step > 0:
            costed_time = time.time() - self.start_time
            predict_all_time = costed_time * max_step / cur_step
            remain_time = predict_all_time - costed_time

            if state.log_history:
                max_loss_entry = max(state.log_history, key=lambda x: x['step'])
                # 获取最大步数对应的 loss 值
                max_loss_value = max_loss_entry.get('loss',-1)
            else:
                max_loss_value = -1
               
        self.cb(self.start_time,costed_time,remain_time,process,max_loss_value)

from opentelemetry import metrics
from opentelemetry.metrics import CallbackOptions,Observation
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter as OTLPMetricGrpcExporter
from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter as OTLPMetricHttpExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
class AgentMetrics():
    def __init__(self,pretrain_name) -> None:
        self.start_train_time = -1
        self.left_train_time = -1
        self.loss = -1
        self.progress = -1
        self.error_code_counter = None
        self.pretrained_model_name = pretrain_name

    def start(self):
        import os
        local_rank = os.environ.get('LOCAL_RANK')
        if local_rank == 0 or local_rank == '0':
            resource = Resource(attributes={
                SERVICE_NAME: "atp-sdk"
            })
            # 使用GRPC协议上报
            task_id = os.environ.get('TRAIN_ID')
            agent_ip = os.environ.get('HostIP')
            if task_id is None or agent_ip is None:
                print('env TRAIN_ID or PRETRAINED_MODEL_NAME None')
                return
            #agent_ip = '127.0.0.1'
            #agent_ip = '172.31.96.145'
            agent_port = 4317

            import socket
            try:
                socket.create_connection((agent_ip, agent_port), timeout=5)
                print(f"{agent_ip}:{agent_port} is reachable.")
            except (socket.timeout, ConnectionError):
                print(f"{agent_ip}:{agent_port} is not reachable.")
                return
    
            reader = PeriodicExportingMetricReader(OTLPMetricGrpcExporter(
                endpoint=f"{agent_ip}:{agent_port}",
                insecure=True
            ),export_interval_millis = 30*1000)

            # 使用HTTP协议上报
            # span_processor = BatchSpanProcessor(OTLPSpanHttpExporter(
            #     endpoint="<endpoint>",
            # ))

            # 创建 MeterProvider
            meter_provider = MeterProvider(resource=resource,metric_readers=[reader])
            metrics.set_meter_provider(meter_provider)

            # 获取 Meter 对象
            meter = metrics.get_meter(__name__)

            def callback_train_progress(options: CallbackOptions):
                if self.left_train_time == -1:
                    return
                report_time = time.time()
                yield Observation(self.progress,{"train_id":task_id,"pretrained_model_name":self.pretrained_model_name,
                                                "left_train_time":self.left_train_time,
                                                "start_train_time":self.start_train_time,
                                                "report_time":report_time})

            def callback_train_loss(options: CallbackOptions):
                if self.loss == -1:
                    return
                report_time = time.time()
                yield Observation(self.loss,{"train_id":task_id,"pretrained_model_name":self.pretrained_model_name,
                                             "report_time":report_time})

            # 创建 Gauge 指标
            gauge_progress = meter.create_observable_gauge(
                name="ATP_TRAIN_PROGRESS",
                description="apt-sdk progress metrics",
                unit="1",
                callbacks=[callback_train_progress]
            )

            gauge_loss = meter.create_observable_gauge(
                name="ATP_TRAIN_LOSS",
                description="apt-sdk loss metrics",
                unit="1",
                callbacks=[callback_train_loss]
            )

            self.error_code_counter = meter.create_counter(
                name="requests",
                description="error code",
            )
            print(f'AgentMetrics start success')

    def set_params(self,start_train_time,left_train_time,loss,progress):
        self.start_train_time = start_train_time
        self.left_train_time = left_train_time
        self.loss = loss
        self.progress = progress

    def set_error_code(self,error_code):
        import os
        import time
        task_id = os.environ.get('TRAIN_ID')

        if task_id is None:
            print('env TRAIN_ID or PRETRAINED_MODEL_NAME None')
            return
        self.error_code_counter.add(error_code,{"train_id":task_id,"pretrained_model_name":self.pretrained_model_name})

    
class HttpReportAgent():
    def __init__(self,pretrain_name) -> None:
        self.start_train_time = -1
        self.left_train_time = -1
        self.loss = -1
        self.progress = -1
        self.error_code_counter = None
        self.pretrained_model_name = pretrain_name
        self.task_id = os.environ.get('TRAIN_ID')
        self.max_retries = 3
        self.url = os.environ.get('METRICS_REPORT_URL')
        self.interval = 30
    
    def make_http_request(self, post_data):
        retries = 0
        while retries < self.max_retries:
            try:
                requests.post(self.url, data=post_data, timeout=5)
                break
            except requests.Timeout:
                retries+=1


    def report_progress(self):
        d = {
            "train_id":self.task_id,
            "pretrained_model_name": self.pretrained_model_name,       
            "left_train_time":self.left_train_time,                                       
            "start_train_time":self.start_train_time,
            "report_time": time.time(),
            "progress": self.progress
        }

        self.make_http_request(d)

    def report_loss(self):
        d = {
            "train_id":self.task_id,
            "pretrained_model_name": self.pretrained_model_name,       
            "report_time": time.time(),
            "loss": self.loss
        }

        self.make_http_request(d)

    def loop_execute_interval(self, fn):
        fn()
        threading.Timer(self.interval, self.loop_execute_interval, (fn)).start()


    def start(self):
        threading.Timer(self.interval, self.loop_execute_interval,(self.report_progress)).start()
        threading.Timer(self.interval, self.loop_execute_interval,(self.report_loss)).start()

        
    def set_params(self,start_train_time,left_train_time,loss,progress):
        self.start_train_time = start_train_time
        self.left_train_time = left_train_time
        self.loss = loss
        self.progress = progress

